package be.iminds.ilabt.jfed.experimenter_gui.gts_editor.views;

import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.rspec.CanvasNodeFactory;
import be.iminds.ilabt.jfed.experimenter_gui.editor.views.EditableCanvasEditorView;
import be.iminds.ilabt.jfed.experimenter_gui.editor.views.EditorViewType;
import be.iminds.ilabt.jfed.experimenter_gui.gts_editor.GTSModelEditor;
import be.iminds.ilabt.jfed.experimenter_gui.gts_editor.toolbox.*;
import be.iminds.ilabt.jfed.groovy_dsl.DSLStreamException;
import be.iminds.ilabt.jfed.gts.GtsDslWriter;
import com.google.common.eventbus.EventBus;
import javafx.application.Platform;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.collections.ListChangeListener;
import javafx.event.EventHandler;
import javafx.geometry.Point2D;
import javafx.scene.Node;
import javafx.scene.control.TitledPane;
import javafx.scene.input.*;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.Region;
import javafx.scene.layout.TilePane;
import javafx.scene.layout.VBox;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

/**
 * User: twalcari
 * Date: 5/19/2015
 * Time: 16:39
 */
public class EditableGTSCanvasEditorView extends BorderPane implements EditableGtsDslView {
    private static final Logger LOG = LoggerFactory.getLogger(EditableGTSCanvasEditorView.class);

    private final GTSModelEditor modelEditor;

    private final EditableGTSCanvas canvas;
    private final VBox toolbox;
    private final TilePane resourcesPane;

    private final StringProperty statusProperty = new SimpleStringProperty();

    private final List<GTSToolboxItem> toolboxItems = new ArrayList<>();

    private final EventHandler<MouseEvent> toolboxOnDragDetectedHandler = mouseEvent -> {
        //set selected toolboxitem as draggedItem
        GTSToolboxItem originalItem = (GTSToolboxItem) mouseEvent.getSource();

        Dragboard dragboard = originalItem.startDragAndDrop(TransferMode.COPY);
        ClipboardContent content = new ClipboardContent();
        content.put(GTSToolboxItem.getDragDataFormat(), originalItem.getId());
        dragboard.setContent(content);
        dragboard.setDragView(originalItem.getDragImage());

        setStatus("Drop the item somewhere in the canvas.");

        mouseEvent.consume();
    };

    private final EventHandler<DragEvent> toolboxOnDragDoneHandler = event -> {

        if (event.getGestureTarget() == null)
            setStatus("Item can only be dropped on the canvas");

        event.consume();
    };


    EditableGTSCanvasEditorView(GTSModelEditor modelEditor, JFedGuiConfig config, EventBus eventBus) {
        this.modelEditor = modelEditor;

        getStylesheets().add(EditableCanvasEditorView.class.getResource("ExperimenterEditor.css").toExternalForm());

        this.canvas = new EditableGTSCanvas(modelEditor, config, eventBus);

        this.canvas.statusProperty().addListener(observable -> {
            setStatus(canvas.getStatus());
        });

        this.canvas.setOnDragOver(event -> {
            if (event.getDragboard().hasContent(GTSToolboxItem.getDragDataFormat())) {
                event.acceptTransferModes(TransferMode.COPY);
            }
            event.consume();
        });

        this.canvas.setOnDragDropped(event -> {
            String draggedItemId =
                    (String) event.getDragboard().getContent(GTSToolboxItem.getDragDataFormat());
            GTSToolboxItem draggedItem = findToolboxItemById(draggedItemId);

            Point2D location = new Point2D(
                    event.getSceneX() - getCanvas().getLocalToSceneTransform().getTx(),
                    event.getSceneY() - getCanvas().getLocalToSceneTransform().getTy()
            );

            //perform this action out-of-band
            Platform.runLater(() -> draggedItem.onAddedToCanvas(canvas, location));

            event.consume();
        });

        setCenter(canvas);

        this.resourcesPane = new TilePane();
        resourcesPane.getStyleClass().add("toolbox-pane");
        resourcesPane.setPrefWidth(200);

        //automatically register drag handlers on all items in the resources pane
        resourcesPane.getChildren().addListener((ListChangeListener<Node>) c -> {
            while (c.next()) {
                if (c.wasAdded()) {
                    for (Node n : c.getAddedSubList()) {
                        n.setOnDragDetected(toolboxOnDragDetectedHandler);
                    }
                }
            }
        });


        TitledPane resourcesTitledPane = new TitledPane();
        resourcesTitledPane.setText("Resources");
        resourcesTitledPane.setContent(resourcesPane);
        resourcesTitledPane.setCollapsible(false);

        this.toolbox = new VBox();
        toolbox.getChildren().add(resourcesTitledPane);

        resourcesPane.getChildren().addListener((ListChangeListener<Node>) c -> {
            while (c.next()) {
                for (Node node : c.getAddedSubList()) {
                    assert (node instanceof GTSToolboxItem);
                    registerToolboxDragHandlers((GTSToolboxItem) node);

                }
            }
        });
        resourcesPane.prefHeightProperty().bind(toolbox.heightProperty());

        setLeft(toolbox);

        //load toolbox content
        loadToolboxContent();
    }

    private GTSToolboxItem findToolboxItemById(String draggedItemId) {
        return toolboxItems.stream()
                .filter(item -> draggedItemId.equals(item.getId())).findFirst()
                .orElseThrow(() -> new RuntimeException("Could not find GTSToolboxItem with id " + draggedItemId));
    }

    private void loadToolboxContent() {
        toolboxItems.add(new GTSHostToolboxItem());
        toolboxItems.add(new GTSPortToolboxItem());

        toolboxItems.add(new GTSOFXToolboxItem());
        toolboxItems.add(new GTSExternalNetworkToolboxItem());
        toolboxItems.add(new GTSTypeToolboxItem());

        toolboxItems.sort(Comparator.naturalOrder());

        resourcesPane.getChildren().addAll(toolboxItems);
    }

    private void registerToolboxDragHandlers(final GTSToolboxItem toolboxItem) {
        toolboxItem.setOnDragDetected(toolboxOnDragDetectedHandler);
        toolboxItem.setOnDragDone(toolboxOnDragDoneHandler);
    }

    public void setStatus(String status) {
        statusProperty.set(status);
    }

    public StringProperty statusProperty() {
        return statusProperty;
    }

    @Override
    public Region getView() {
        return this;
    }

    @Override
    public EditorViewType getType() {
        return EditorViewType.CANVAS;
    }

    @Override
    public String getResultDsl() {
        GtsDslWriter writer = new GtsDslWriter(modelEditor.getModel());

        try {
            return writer.call();
        } catch (DSLStreamException e) {
            LOG.error("Error while writing GTS DSL", e);
            throw new RuntimeException("Error while writing GTS DSL", e);
        }
    }


    @Override
    public void dispose() {

    }


    public EditableGTSCanvas getCanvas() {
        return canvas;
    }

    public GTSModelEditor getModelEditor() {
        return modelEditor;
    }
}
